"""
ExplaiNN
interpretable and transparent neural networks for genomics
"""

from .utils import tools
from .interpretation import interpretation
from .train import train, test
from .models import networks

__version__ = "0.1.5"


# =============================================================================
# IMPORTS
# =============================================================================
from collections import OrderedDict

import torch.nn.functional as F
import torch.utils.data
import torch.nn as nn
import torch.nn.modules.activation as activation
from explainn.utils.tools import _flip

# =============================================================================
# MODEL CLASSES
# =============================================================================
class exp_activation_module(nn.Module):
    """
    Exponential activation function
    (from Koo, 2021, PMID: 34322657)
    """
    def __init__(self):
        super(exp_activation_module, self).__init__()
    def forward(self, x):
        return torch.exp(x)

class Unsqueeze(torch.nn.Module):
    """
    Unsqueeze for sequential models
    """
    def forward(self, x):
        return x.unsqueeze(-1)

class ConvNetDeep(nn.Module):
    """
    CNN with 3 conv layers;
    Inspired by Basset (PMID: 27197224)
    """
    def __init__(self, num_classes, weight_path=None):
        """
        initialize the model
        designed for input sequences of length 200 bp

        :param num_classes: int, number of outputs
        :param weight_path: string, path to the file with model weights
        """
        super(ConvNetDeep, self).__init__()
        # Block 1 :
        self.c1 = nn.Conv1d(4, 100, 19)
        self.bn1 = nn.BatchNorm1d(100)
        self.rl1 = activation.ReLU()
        self.mp1 = nn.MaxPool1d(3, 3)

        # Block 2 :
        self.c2 = nn.Conv1d(100, 200, 7)
        self.bn2 = nn.BatchNorm1d(200)
        self.rl2 = activation.ReLU()
        self.mp2 = nn.MaxPool1d(3, 3)

        # Block 3 :
        self.c3 = nn.Conv1d(200, 200, 4)
        self.bn3 = nn.BatchNorm1d(200)
        self.rl3 = activation.ReLU()
        self.mp3 = nn.MaxPool1d(3, 3)

        # Block 4 : Fully Connected 1 :
        self.d4 = nn.Linear(1000, 1000)  # 1000 for 200 input size
        self.bn4 = nn.BatchNorm1d(1000, 1e-05, 0.1, True)
        self.rl4 = activation.ReLU()
        self.dr4 = nn.Dropout(0.3)

        # Block 5 : Fully Connected 2 :
        self.d5 = nn.Linear(1000, 1000)
        self.bn5 = nn.BatchNorm1d(1000, 1e-05, 0.1, True)
        self.rl5 = activation.ReLU()
        self.dr5 = nn.Dropout(0.3)

        # Block 6 :4Fully connected 3
        self.d6 = nn.Linear(1000, num_classes)
        # self.sig = activation.Sigmoid()

        if weight_path:
            self.load_weights(weight_path)

    def forward(self, x, embeddings=False):
        """
            :param: embeddings : if True forward return embeddings along with the output
        """
        x = self.rl1(self.bn1(self.c1(x)))

        activations = x
        x = self.mp1(x)
        x = self.mp2(self.rl2(self.bn2(self.c2(x))))
        em = self.mp3(self.rl3(self.bn3(self.c3(x))))
        o = torch.flatten(em, start_dim=1)
        o = self.dr4(self.rl4(self.bn4(self.d4(o))))
        o = self.dr5(self.rl5(self.bn5(self.d5(o))))
        o = self.d6(o)

        activations, act_index = torch.max(activations, dim=2)

        if embeddings: return (o, activations, act_index, em)
        return o

    def load_weights(self, weight_path):
        sd = torch.load(weight_path)
        new_dict = OrderedDict()
        keys = list(self.state_dict().keys())
        values = list(sd.values())
        for i in range(len(values)):
            v = values[i]
            if v.dim() > 1:
                if v.shape[-1] == 1:
                    new_dict[keys[i]] = v.squeeze(-1)
                    continue
            new_dict[keys[i]] = v
        self.load_state_dict(new_dict)


#################################################################################
class ConvNetShallow(nn.Module):
    """
    Shallow CNN model with 1 convolutional layer
    """
    def __init__(self, num_classes, weight_path=None):
        """
        initialize the model
        designed for input sequences of length 200 bp

        :param num_classes: int, number of output classes
        :param weight_path: string, path to the file with model weights
        """
        super(ConvNetShallow, self).__init__()
        # Block 1 :
        self.c1 = nn.Conv1d(4, 100, 19)
        self.bn1 = nn.BatchNorm1d(100)
        self.rl1 = activation.ReLU()
        self.mp1 = nn.MaxPool1d(7, 7)

        # Block 4 : Fully Connected 1 :
        self.d2 = nn.Linear(2600, 1000)  # 1000 for 200 input size
        self.bn2 = nn.BatchNorm1d(1000, 1e-05, 0.1, True)
        self.rl2 = activation.ReLU()
        self.dr2 = nn.Dropout(0.3)

        # Block 5 : Fully Connected 2 :
        self.d3 = nn.Linear(1000, 1000)
        self.bn3 = nn.BatchNorm1d(1000, 1e-05, 0.1, True)
        self.rl3 = activation.ReLU()
        self.dr3 = nn.Dropout(0.3)

        # Block 6 :4Fully connected 3
        self.d4 = nn.Linear(1000, num_classes)
        # self.sig = activation.Sigmoid()

        if weight_path:
            self.load_weights(weight_path)

    def forward(self, x, embeddings=False):
        """
            :param: embeddings : if True forward return embeddings along with the output
        """
        x = self.rl1(self.bn1(self.c1(x)))

        activations = x
        em = self.mp1(x)
        o = torch.flatten(em, start_dim=1)
        o = self.dr2(self.rl2(self.bn2(self.d2(o))))
        o = self.dr3(self.rl3(self.bn3(self.d3(o))))
        o = self.d4(o)

        activations, act_index = torch.max(activations, dim=2)

        if embeddings: return (o, activations, act_index, em)
        return o

    def load_weights(self, weight_path):
        sd = torch.load(weight_path)
        new_dict = OrderedDict()
        keys = list(self.state_dict().keys())
        values = list(sd.values())
        for i in range(len(values)):
            v = values[i]
            if v.dim() > 1:
                if v.shape[-1] == 1:
                    new_dict[keys[i]] = v.squeeze(-1)
                    continue
            new_dict[keys[i]] = v
        self.load_state_dict(new_dict)


#################################################################################
class DanQ(nn.Module):
    """
    Implementation of DanQ architecture
    (PMID: 27084946)
    """
    def __init__(self, num_classes, weight_path=None):
        """
        initialize the model
        designed for input sequences of length 200 bp

        :param num_classes: int, number of output classes
        :param weight_path: string, path to the file with model weights
        """
        super(DanQ, self).__init__()
        self.Conv1 = nn.Conv1d(in_channels=4, out_channels=320, kernel_size=26)
        self.Maxpool = nn.MaxPool1d(kernel_size=13, stride=13)
        self.Drop1 = nn.Dropout(p=0.2)
        self.BiLSTM = nn.LSTM(input_size=320, hidden_size=320, num_layers=2,
                              batch_first=True,
                              dropout=0.5,
                              bidirectional=True)
        self.Linear1 = nn.Linear(13 * 640, 925)
        self.Linear2 = nn.Linear(925, num_classes)

        if weight_path:
            self.load_weights(weight_path)

    def forward(self, input):
        x = self.Conv1(input)
        x = F.relu(x)
        x = self.Maxpool(x)
        x = self.Drop1(x)
        x_x = torch.transpose(x, 1, 2)
        x, (h_n, h_c) = self.BiLSTM(x_x)
        x = x.contiguous().view(-1, 13 * 640)
        x = self.Linear1(x)
        x = F.relu(x)
        x = self.Linear2(x)
        return x

    def load_weights(self, weight_path):
        sd = torch.load(weight_path)
        new_dict = OrderedDict()
        keys = list(self.state_dict().keys())
        values = list(sd.values())
        for i in range(len(values)):
            v = values[i]
            if v.dim() > 1:
                if v.shape[-1] == 1:
                    new_dict[keys[i]] = v.squeeze(-1)
                    continue
            new_dict[keys[i]] = v
        self.load_state_dict(new_dict)


class ExplaiNN(nn.Module):
    """
    Class for the ExplaiNN model
    """
    def __init__(self, num_cnns, input_length, num_classes, filter_size=19, num_fc=2, pool_size=7, pool_stride=7,
                 weight_path=None):
        """
        initialize the model

        :param num_cnns: int, number of independent cnn units
        :param input_length: int, input sequence length
        :param num_classes: int, number of outputs
        :param filter_size: int, size of the unit's filter, default=19
        :param num_fc: int, number of FC layers in the unit, default=2
        :param pool_size: int, size of the unit's maxpooling layer, default=7
        :param pool_stride: int, stride of the unit's maxpooling layer, default=7
        :param weight_path: string, path to the file with model weights
        """
        super(ExplaiNN, self).__init__()

        if num_fc == 0:
            self.linears = nn.Sequential(
                nn.Conv1d(in_channels=4 * num_cnns, out_channels=1 * num_cnns, kernel_size=filter_size,
                          groups=num_cnns),
                nn.BatchNorm1d(num_cnns),
                exp_activation_module(),
                nn.MaxPool1d(input_length - (filter_size-1)),
                nn.Flatten())
        elif num_fc == 1:
            self.linears = nn.Sequential(
                nn.Conv1d(in_channels=4 * num_cnns, out_channels=1 * num_cnns, kernel_size=filter_size,
                          groups=num_cnns),
                nn.BatchNorm1d(num_cnns),
                exp_activation_module(),
                nn.MaxPool1d(pool_size, pool_stride),
                nn.Flatten(),
                Unsqueeze(),
                nn.Conv1d(in_channels=int(((input_length - (filter_size-1)) - (pool_size-1)-1)/pool_stride + 1) * num_cnns,
                          out_channels=1 * num_cnns, kernel_size=1,
                          groups=num_cnns),
                nn.BatchNorm1d(1 * num_cnns, 1e-05, 0.1, True),
                nn.ReLU(),
                nn.Flatten())
        elif num_fc == 2:
            self.linears = nn.Sequential(
                nn.Conv1d(in_channels=4 * num_cnns, out_channels=1 * num_cnns, kernel_size=filter_size,
                          groups=num_cnns),
                nn.BatchNorm1d(num_cnns),
                exp_activation_module(),
                nn.MaxPool1d(pool_size, pool_stride),
                nn.Flatten(),
                Unsqueeze(),
                nn.Conv1d(in_channels=int(((input_length - (filter_size-1)) - (pool_size-1)-1)/pool_stride + 1) * num_cnns,
                          out_channels=100 * num_cnns, kernel_size=1,
                          groups=num_cnns),
                nn.BatchNorm1d(100 * num_cnns, 1e-05, 0.1, True),
                nn.ReLU(),
                nn.Dropout(0.3),
                nn.Conv1d(in_channels=100 * num_cnns,
                          out_channels=1 * num_cnns, kernel_size=1,
                          groups=num_cnns),
                nn.BatchNorm1d(1 * num_cnns, 1e-05, 0.1, True),
                nn.ReLU(),
                nn.Flatten())
        else:
            self.linears = nn.Sequential(
                nn.Conv1d(in_channels=4 * num_cnns, out_channels=1 * num_cnns, kernel_size=filter_size,
                          groups=num_cnns),
                nn.BatchNorm1d(num_cnns),
                exp_activation_module(),
                nn.MaxPool1d(pool_size, pool_stride),
                nn.Flatten(),
                Unsqueeze(),
                nn.Conv1d(in_channels=int(((input_length - (filter_size-1)) - (pool_size-1)-1)/pool_stride + 1) * num_cnns,
                          out_channels=100 * num_cnns, kernel_size=1,
                          groups=num_cnns),
                nn.BatchNorm1d(100 * num_cnns, 1e-05, 0.1, True),
                nn.ReLU())

            self.linears_bg = nn.ModuleList([nn.Sequential(nn.Dropout(0.3),
                                                           nn.Conv1d(in_channels=100 * num_cnns,
                                                                     out_channels=100 * num_cnns, kernel_size=1,
                                                                     groups=num_cnns),
                                                           nn.BatchNorm1d(100 * num_cnns, 1e-05, 0.1, True),
                                                           nn.ReLU()) for i in range(num_fc - 2)])

            self.last_linear = nn.Sequential(nn.Dropout(0.3),
                                             nn.Conv1d(in_channels=100 * num_cnns, out_channels=1 * num_cnns,
                                                       kernel_size=1,
                                                       groups=num_cnns),
                                             nn.BatchNorm1d(1 * num_cnns, 1e-05, 0.1, True),
                                             nn.ReLU(),
                                             nn.Flatten())

        self.final = nn.Linear(num_cnns, num_classes)
        self.num_cnns = num_cnns
        self.num_fc = num_fc

        if weight_path:
            self.load_weights(weight_path)

    def forward(self, x):
        x = x.repeat(1, self.num_cnns, 1)

        if self.num_fc <= 2:
            outs = self.linears(x)
        else:
            outs = self.linears(x)
            for i in range(len(self.linears_bg)):
                outs = self.linears_bg[i](outs)
            outs = self.last_linear(outs)

        out = self.final(outs)

        return out

    def load_weights(self, weight_path):
        sd = torch.load(weight_path)
        new_dict = OrderedDict()
        keys = list(self.state_dict().keys())
        values = list(sd.values())
        for i in range(len(values)):
            v = values[i]
            if v.dim() > 1:
                if v.shape[-1] == 1:
                    new_dict[keys[i]] = v.squeeze(-1)
                    continue
            new_dict[keys[i]] = v
        self.load_state_dict(new_dict)


class SingleLayer(nn.Module):
    """
    Single output layer;
    Useful when predicting on the outputs of other models
    """

    def __init__(self, num_inputs, num_classes, weight_path=None):
        """
        initialize the layer

        :param num_inputs: int, number of input features
        :param num_classes: int, number of outputs
        :param weight_path: string, path to the file with weights
        """
        super(SingleLayer, self).__init__()

        self.final = nn.Linear(num_inputs, num_classes)

        if weight_path:
            self.load_weights(weight_path)

    def forward(self, x):

        out = self.final(x.float())

        return out

    def load_weights(self, weight_path):
        sd = torch.load(weight_path)
        new_dict = OrderedDict()
        keys = list(self.state_dict().keys())
        values = list(sd.values())
        for i in range(len(values)):
            v = values[i]
            if v.dim() > 1:
                if v.shape[-1] == 1:
                    new_dict[keys[i]] = v.squeeze(-1)
                    continue
            new_dict[keys[i]] = v
        self.load_state_dict(new_dict)


class NAMLayer(nn.Module):
    """
    Same as the linear class, but NAM style
    (extra FC layers before the final output)
    """
    def __init__(self, num_inputs, num_hidden, num_classes, weight_path=None):
        """
        initialize the model

        :param num_inputs: int, number of inputs
        :param num_hidden: int, size of the FC hidden layer
        :param num_classes: int, number of output classes
        :param weight_path: string, path to file with weights
        """
        super(NAMLayer, self).__init__()

        self.linear = nn.Sequential(
            nn.Conv1d(in_channels=1 * num_inputs, out_channels=num_hidden * num_inputs,
                      kernel_size=1,
                      groups=num_inputs),
            nn.BatchNorm1d(num_hidden * num_inputs, 1e-05, 0.1, True),
            nn.ReLU(),
            nn.Dropout(0.3),
            # nn.Conv1d(in_channels=num_hidden * num_inputs, out_channels=num_hidden * num_inputs,
            #          kernel_size=1,
            #          groups=num_inputs),
            # nn.BatchNorm1d(num_hidden*num_inputs,1e-05,0.1,True),
            # nn.ReLU(),
            # nn.Dropout(0.3),
            nn.Conv1d(in_channels=num_hidden * num_inputs, out_channels=1 * num_inputs,
                      kernel_size=1,
                      groups=num_inputs),
            nn.Flatten()
        )

        self.final = nn.Linear(num_inputs, num_classes)

        if weight_path:
            self.load_weights(weight_path)

    def forward(self, x):

        x = x.unsqueeze(-1)
        xout = self.linear(x)
        out = self.final(xout)

        return out

    def load_weights(self, weight_path):
        sd = torch.load(weight_path)
        new_dict = OrderedDict()
        keys = list(self.state_dict().keys())
        values = list(sd.values())
        for i in range(len(values)):
            v = values[i]
            if v.dim() > 1:
                if v.shape[-1] == 1:
                    new_dict[keys[i]] = v.squeeze(-1)
                    continue
            new_dict[keys[i]] = v
        self.load_state_dict(new_dict)

class DeepSTARR(nn.Module):
    """
    Model taken from Almeida et al
    (PMID: 35551305)
    """
    def __init__(self, weight_path=None):
        super(DeepSTARR, self).__init__()
        self.convol = nn.Sequential(
            nn.Conv1d(in_channels=4, out_channels=256, kernel_size=7,
                      padding=int((7 - 1) / 2)),  # same padding
            nn.BatchNorm1d(256),
            nn.ReLU(),
            nn.MaxPool1d(2, 2),
            nn.Conv1d(in_channels=256, out_channels=60, kernel_size=3,
                      padding=int((3 - 1) / 2)),  # same padding
            nn.BatchNorm1d(60),
            nn.ReLU(),
            nn.MaxPool1d(2, 2),
            nn.Conv1d(in_channels=60, out_channels=60, kernel_size=5,
                      padding=int((5 - 1) / 2)),  # same padding
            nn.BatchNorm1d(60),
            nn.ReLU(),
            nn.MaxPool1d(2, 2),
            nn.Conv1d(in_channels=60, out_channels=120, kernel_size=3,
                      padding=int((3 - 1) / 2)),  # same padding
            nn.BatchNorm1d(120),
            nn.ReLU(),
            nn.MaxPool1d(2, 2),
            nn.Flatten()
        )

        self.linear = nn.Sequential(
            nn.Linear(1800, 256),
            nn.BatchNorm1d(256),
            nn.ReLU(),
            nn.Dropout(0.4),
            nn.Linear(256, 256),
            nn.BatchNorm1d(256),
            nn.ReLU(),
            nn.Dropout(0.4),
            nn.Linear(256, 2)
        )

        if weight_path:
            self.load_weights(weight_path)

    def forward(self, x):

        o = self.linear(self.convol(x))

        return o

    def load_weights(self, weight_path):
        sd = torch.load(weight_path)
        new_dict = OrderedDict()
        keys = list(self.state_dict().keys())
        values = list(sd.values())
        for i in range(len(values)):
            v = values[i]
            if v.dim() > 1:
                if v.shape[-1] == 1:
                    new_dict[keys[i]] = v.squeeze(-1)
                    continue
            new_dict[keys[i]] = v
        self.load_state_dict(new_dict)

class NonStrandSpecific(nn.Module):
    '''
    Adapted from Selene:
    https://github.com/FunctionLab/selene/blob/master/selene_sdk/utils/non_strand_specific_module.py

    A torch.nn.Module that wraps a user-specified model architecture if the
    architecture does not need to account for sequence strand-specificity.

    Parameters
    ----------
    model : torch.nn.Module
        The user-specified model architecture.
    mode : {'mean', 'max'}, optional
        Default is 'mean'. NonStrandSpecific will pass the input and the
        reverse-complement of the input into `model`. The mode specifies
        whether we should output the mean or max of the predictions as
        the non-strand specific prediction.

    Attributes
    ----------
    model : torch.nn.Module
        The user-specified model architecture.
    mode : {'mean', 'max'}
        How to handle outputting a non-strand specific prediction.
    '''

    def __init__(self, model):
        super(NonStrandSpecific, self).__init__()

        self.model = model

    def forward(self, input):
        reverse_input = None
        reverse_input = _flip(_flip(input, 1), 2)

        output = self.model.forward(input)
        output_from_rev = self.model.forward(reverse_input)

        return((output + output_from_rev) / 2)


# =============================================================================
# IMPORTS
# =============================================================================
import torch
import copy

# =============================================================================
# FUNCTIONS
# =============================================================================

def train_explainn(train_loader, test_loader, model, device, criterion, optimizer, num_epochs,
                          weights_folder, name_ind, verbose, trim_weights, checkpoint=1):
    """
    Function to train the ExplaiNN model

    :param train_loader: pytorch DataLoader, train data
    :param test_loader: pytorch DataLoader, validation data
    :param model: ExplaiNN model
    :param device: current available device ('cuda:0' or 'cpu')
    :param criterion: objective (loss) function to use (e.g. MSELoss)
    :param optimizer: pytorch Optimizer (e.g. SGD)
    :param num_epochs: int, number of epochs to train the model
    :param weights_folder: string, folder where to save checkpoints
    :param name_ind: string, suffix name of the checkpoints
    :param verbose: boolean, if False, does not print the progress
    :param trim_weights: boolean, if True, makes output layer weights non negative
    :param checkpoint: int, how often to save checkpoints. For example, 1 means that every model is saved
    :return: tuple:
                    trained ExplaiNN model,
                    list, train losses,
                    list, test losses
    """

    train_error = []
    test_error = []

    best_model_wts = copy.deepcopy(model.state_dict())
    best_loss_valid = float('inf')
    best_epoch = 1

    for epoch in range(1,num_epochs+1):

        running_loss = 0.0

        model.train()
        for seqs, labels in train_loader:
            x = seqs.to(device)
            labels = labels.to(device)

            optimizer.zero_grad()

            outputs = model(x)
            loss = criterion(outputs, labels)

            # Backward and optimize
            loss.backward()

            optimizer.step()

            # to clip the weights (constrain them to be non-negative)
            if trim_weights:
                model.final.weight.data.clamp_(0)

            running_loss += loss.item()

        # save training loss to file
        epoch_loss = running_loss / len(train_loader)
        train_error.append(epoch_loss)

        # calculate test (validation) loss for epoch
        test_loss = 0.0

        with torch.no_grad():  # we don't train and don't save gradients here
            model.eval()  # we set forward module to change dropout and batch normalization techniques
            for seqs, labels in test_loader:
                x = seqs.to(device)
                y = labels.to(device)
                outputs = model(x)
                loss = criterion(outputs, y)
                test_loss += loss.item()

        test_loss = test_loss / len(test_loader)
        test_error.append(test_loss)

        if verbose:
            print('Epoch [{}], Current Train Loss: {:.5f}, Current Val Loss: {:.5f}'
                  .format(epoch, epoch_loss, test_loss))

        if test_loss < best_loss_valid:
            best_loss_valid = test_loss
            best_epoch = epoch
            best_model_wts = copy.deepcopy(model.state_dict())

        if epoch % checkpoint == 0:
            model.load_state_dict(best_model_wts)
            torch.save(best_model_wts, weights_folder + "/"+"model_epoch_"+str(epoch)+"_"+
                          name_ind+".pth")

    model.load_state_dict(best_model_wts)
    torch.save(best_model_wts, weights_folder + "/" + "model_epoch_best_" + str(best_epoch) + "_" +
               name_ind + ".pth")  # weights_folder, name_ind

    return model, train_error, test_error


# =============================================================================
# IMPORTS
# =============================================================================
import torch.nn as nn
import pandas as pd
import torch
import numpy as np
from tqdm import tqdm

# =============================================================================
# FUNCTIONS
# =============================================================================
def get_average_activ(self, inputs, outputs):
    """
    Pytorch Hook that will get the average activation for each layer on the current batch
    which can be usefull to get the average usage of the filter
    :param self: pytorch layer, the layer the hook is attached to
    :param inputs: tensor, current input tensor of the layer
    :param outputs: tensor, current output tensor of the layer
    """

    if self.mode[0] == 'Baseline':
        self.register_buffer('average_activation', outputs.mean(0).mean(1))


# hook to assign the filter values to the average activation across batches
def set_filt_to_aver(self, inputs, outputs):
    """
    Pytorch hook to assign the filter values to the average activation across batches
    :param self: pytorch layer, the layer the hook is attached to
    :param inputs: tensor, current input tensor of the layer
    :param outputs: tensor, current output tensor of the layer
    """

    if self.mode[0] == 'Compare':
        outputs[:, self.mode[1], :] = self.average_activation[self.mode[1]]


def nullify_filter_strict(self, input, output):
    """
    Pytorch Hook that will nullify the output of one of the filter indicated in mode for that layer
    :param self: pytorch layer, the layer the hook is attached to
    :param input: tensor, current input tensor of the layer
    :param output: tensor, current output tensor of the layer
    """

    if self.mode[0] == 'Compare':
        output[:, self.mode[1], :] = 0

def get_explainn_predictions(data_loader, model, device, isSigmoid=False):
    """
    Function to get predictions and true labels for all sequences in the given data loader
    :param data_loader: torch DataLoader, data loader with the sequences of interest
    :param model: ExplaiNN model
    :param device: current available device ('cuda:0' or 'cpu')
    :param isSigmoid: boolean, True if the model output is binary
    :return: tuple of numpy.arrays, model predictions and true labels
    """

    running_outputs = []
    running_labels = []

    with torch.no_grad():
        for seq, lbl in data_loader:
            seq = seq.to(device)
            out = model(seq)
            if isSigmoid:
                sigmoid = nn.Sigmoid()
                out = sigmoid(out.detach().cpu())
            else:
                out = out.detach().cpu()
            running_outputs.extend(out.numpy())
            running_labels.extend(lbl.numpy())

    running_labels = np.array(running_labels)
    running_outputs = np.array(running_outputs)

    return running_outputs, running_labels

#
def get_explainn_unit_outputs(data_loader, model, device):
    """
    Function to get predictions and true labels for all sequences in the given data loader
    :param data_loader: torch DataLoader, data loader with the sequences of interest
    :param model: ExplaiNN model
    :param device: current available device ('cuda:0' or 'cpu')
    :return: numpy.array, outputs of individual units, shape (N, U); N - size of the dataset; U - number of units;
    """

    running_cnn_outputs = []
    for seq, lbl in data_loader:
        x = seq.to(device)
        x = x.repeat(1, model.num_cnns, 1)
        cnn_output = model.linears(x)
        running_cnn_outputs.extend(cnn_output.cpu().detach().numpy())

    running_cnn_outputs = np.array(running_cnn_outputs)
    return running_cnn_outputs


#????? can be applied to SingleLayer??
# def get_namlayer_unit_activations(data_loader, model, device):
#     """
#     Function to scan input sequences by NAMLayer model and compute the individual node outputs
#     (activations)
#     :param data_loader: torch DataLoader, the sequence dataset
#     :param model: NAMLayer model
#     :param device: current available device ('cuda:0' or 'cpu')
#     :return: numpy.array, matrix of activations of shape (N, U); N - size of the dataset; U - number of units
#     """
#
#     running_cnn_outputs = []
#
#     for seq, lbl in data_loader:
#         x = seq.to(device)
#
#         x = x.unsqueeze(-1)
#         cnn_output = model.linear(x)
#
#         running_cnn_outputs.extend(cnn_output.cpu().detach().numpy())
#     running_cnn_outputs = np.array(running_cnn_outputs)
#     return running_cnn_outputs

def get_explainn_unit_activations(data_loader, model, device):
    """
    Function to scan input sequences by ExplaiNN model convolutional filters and compute the convolutional units outputs
    (activations)
    :param data_loader: torch DataLoader, the sequence dataset
    :param model: ExplaiNN model
    :param device: current available device ('cuda:0' or 'cpu')
    :return: numpy.array, matrix of activations of shape (N, U, S); N - size of the dataset; U - number of units;
    S - size of the activation map
    """

    running_activations = []

    with torch.no_grad():
        for seq, lbl in tqdm(data_loader, total=len(data_loader)):
            seq = seq.to(device)

            seq = seq.repeat(1, model.num_cnns, 1)
            act = model.linears[:3](seq)

            running_activations.extend(act.cpu().numpy())

    return np.array(running_activations)


def get_pwms_explainn(activations, sequences, y, filter_size):
    """
    Function to convert filter activation values to PWMs
    :param activations: numpy.array, matrix of activations of shape (N, U, S); N - size of the dataset;
    U - number of units; S - size of the activation map
    :param sequences: torch, array of one-hot encoded sequences
    :param y: torch, array of output labels
    :param filter_size: int, size of the filter
    :return: numpy.array, pwm matrices, shape (U, 4, filter_size), where U - number of units
    """

    # find the threshold value for activation
    activation_threshold = 0.5 * np.amax(activations, axis=(0, 2))

    # Get the number of units/filters
    n_filters = activations.shape[1]

    # pad sequences:
    # npad = ((0, 0), (0, 0), (9, 9))
    # sequences = np.pad(sequences, pad_width=npad, mode='constant', constant_values=0)

    pwm = np.zeros((n_filters, 4, filter_size))
    # pfm = np.zeros((n_filters, 4, filter_size))
    n_samples = activations.shape[0]

    activation_indices = []

    for i in tqdm(range(n_filters)):
        # create list to store filter_size bp sequences that activated filter
        act_seqs_list = []

        for j in range(n_samples):
            # find all indices where filter is activated
            indices = np.where(activations[j, i, :] > activation_threshold[i])

            for start in indices[0]:
                activation_indices.append(start)
                end = start + filter_size
                act_seqs_list.append(sequences[j, :, start:end])

        # convert act_seqs from list to array
        if act_seqs_list:
            act_seqs = np.stack(act_seqs_list)
            pwm_tmp = np.sum(act_seqs, axis=0)
            # pfm_tmp = pwm_tmp
            total = np.sum(pwm_tmp, axis=0)
            pwm_tmp = np.nan_to_num(pwm_tmp / total)

            pwm[i] = pwm_tmp
            # pfm[i] = pfm_tmp

    return pwm


def pwm_to_meme(pwm, output_file_path):
    """
    Function to convert pwm array to meme file
    :param pwm: numpy.array, pwm matrices, shape (U, 4, filter_size), where U - number of units
    :param output_file_path: string, the name of the output meme file
    """

    n_filters = pwm.shape[0]
    filter_size = pwm.shape[2]
    meme_file = open(output_file_path, 'w')
    meme_file.write("MEME version 4 \n")

    print('Saved PWM File as : {}'.format(output_file_path))

    for i in range(0, n_filters):
        if np.sum(pwm[i, :, :]) > 0:
            meme_file.write("\n")
            meme_file.write("MOTIF filter%s \n" % i)
            meme_file.write(
                "letter-probability matrix: alength= 4 w= %d \n" % np.count_nonzero(np.sum(pwm[i, :, :], axis=0)))

        for j in range(0, filter_size):
            if np.sum(pwm[i, :, j]) > 0:
                meme_file.write(str(pwm[i, 0, j]) + "\t" + str(pwm[i, 1, j]) + "\t" + str(pwm[i, 2, j]) + "\t" + str(
                    pwm[i, 3, j]) + "\n")

    meme_file.close()

def get_median_unit_importance(activations, model, unit_outputs, target_labels):
    """
    Function to compute median unit importance (unit_output*class weight) of ExplaiNN units
    :param activations: numpy.array, matrix of activations of shape (N, U, S); N - size of the dataset;
    U - number of units; S - size of the activation map
    :param model: ExplaiNN model
    :param unit_outputs: numpy.array, outputs of individual units, shape (N, U); N - size of the dataset; U - number of units;
    :param target_labels: a list with the names of the output nodes;
    :return: numpy.array, a matrix of size (U, O); U - number of units; O - number of labels/classes;
    """

    activation_threshold = 0.5 * np.amax(activations, axis=(0, 2))

    # sequences (their indeces) that highly activated the filter
    res = {}
    for i in range(100):
        inds = []
        for j in range(activations.shape[0]):
            indices = np.where(activations[j, i, :] > activation_threshold[i])
            if indices[0].shape[0] > 0:
                inds.append(j)
        res[i] = inds

    weights = model.final.weight.detach().cpu().numpy()  # -0.035227 0.480355

    feat_imp_median = np.zeros((100, 50))

    for filt in tqdm(range(100)):
        res_distr = {}
        for cl in range(len(target_labels)):
            f_cell = np.multiply(unit_outputs, weights[cl])
            res_distr[target_labels[cl]] = f_cell[:, filt]
            res_distr[target_labels[cl]] = res_distr[target_labels[cl]][res[filt]]
        res_distr = pd.Series(res_distr)

        res_distr = res_distr.apply(lambda x: np.median(x))
        feat_imp_median[filt, :] = res_distr.values

    return feat_imp_median


def get_specific_unit_importance(activations, model, unit_outputs, filt, target_labels):
    """
    Function to compute unit importance (unit_output*class weight) of a particular ExplaiNN unit (indexed at filt)
    :param activations: numpy.array, matrix of activations of shape (N, U, S); N - size of the dataset;
    U - number of units; S - size of the activation map
    :param model: ExplaiNN model
    :param unit_outputs: numpy.array, outputs of individual units, shape (N, U); N - size of the dataset; U - number of units;
    :param filt: int, index of the unit of interest;
    :param target_labels: a list with the names of the output nodes;
    :return: pandas.Series, contains O keys (number of ExplaiNN outputs, labels), each key contains an array of size X,
    where X is equal to the number of sequences that activated the unit of interest (indexed at filt) more than an
    activation threshold
    """

    activation_threshold = 0.5 * np.amax(activations, axis=(0, 2))

    # sequences (their indeces) that highly activated the filter
    res = {}
    for i in range(100):
        inds = []
        for j in range(activations.shape[0]):
            indices = np.where(activations[j, i, :] > activation_threshold[i])
            if indices[0].shape[0] > 0:
                inds.append(j)
        res[i] = inds

    weights = model.final.weight.detach().cpu().numpy()  # -0.035227 0.480355

    res_distr = {}
    for cl in range(len(target_labels)):
        f_cell = np.multiply(unit_outputs, weights[cl])
        res_distr[target_labels[cl]] = f_cell[:, filt]
        res_distr[target_labels[cl]] = res_distr[target_labels[cl]][res[filt]]

    res_distr = pd.Series(res_distr)

    return res_distr


# =============================================================================
# IMPORTS
# =============================================================================
import torch.nn as nn
import torch
import numpy as np

# =============================================================================
# FUNCTIONS
# =============================================================================

def run_test(model, data_loader, device, isSigmoid=False):
    """

    :param model: ExplaiNN model
    :param data_loader: torch DataLoader, data loader with the sequences of interest
    :param device: current available device ('cuda:0' or 'cpu')
    :param isSigmoid: boolean, True if the model output is binary
    :return:
    """
    running_outputs = []
    running_labels = []
    sigmoid = nn.Sigmoid()
    with torch.no_grad():
        for seq, lbl in data_loader:
            seq = seq.to(device)
            out = model(seq)
            out = out.detach().cpu()
            if isSigmoid:
                out = sigmoid(out)
            running_outputs.extend(out.numpy())
            running_labels.extend(lbl.numpy())
    return np.array(running_labels), np.array(running_outputs)


# =============================================================================
# IMPORTS
# =============================================================================
import torch
import numpy as np
import h5py
from tqdm import tqdm
import torch.nn as nn
import matplotlib.pyplot as plt

# =============================================================================
# FUNCTIONS
# =============================================================================
def count_parameters(model):
    """
    Calculates the number of parameters in the model

    :param model: pytorch model
    :return: int, number of parameters
    """

    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def showPlot(points, points2, title, save=None):
    """
    function to show training curve
    save - place to save the figure
    :param points: list, training loss
    :param points2: list, validation loss
    :param title: string, optional title
    :param save: boolean, save figure as a sep file or not
    :return:
    """

    plt.figure()
    fig, ax = plt.subplots()
    # this locator puts ticks at regular intervals
    # loc = ticker.MultipleLocator(base=0.2)
    # ax.yaxis.set_major_locator(loc)
    plt.plot(points)
    plt.plot(points2)
    plt.ylabel("Loss")
    plt.legend(['train', 'validation'], loc='upper right')
    plt.title(title)

    if save:
        plt.savefig(save)
    else:
        plt.show()


def load_datas(path_h5, batch_size, num_workers, shuffle):
    """
    Loads the dataset from an h5 file

    :param path_h5: string, path to the file
    :param batch_size: int, batch size to use
    :param num_workers: int, number of workers for DataLoader
    :param shuffle: boolean, shuffle parameter in DataLoader
    :return: tuple:
        a dictionary with train, validation, and test sets;
        a list with the names of the output nodes;
        an array with true values for the train samples;
    """

    data = h5py.File(path_h5, 'r')
    dataset = {}
    dataloaders = {}
    # Train data
    dataset['train'] = torch.utils.data.TensorDataset(torch.Tensor(np.array(data['train_in'])),
                                                      torch.Tensor(np.array(data['train_out'])))
    dataloaders['train'] = torch.utils.data.DataLoader(dataset['train'],
                                                       batch_size=batch_size, shuffle=shuffle,
                                                       num_workers=num_workers)

    # Validation data
    dataset['valid'] = torch.utils.data.TensorDataset(torch.Tensor(np.array(data['valid_in'])),
                                                      torch.Tensor(np.array(data['valid_out'])))
    dataloaders['valid'] = torch.utils.data.DataLoader(dataset['valid'],
                                                       batch_size=batch_size, shuffle=shuffle,
                                                       num_workers=num_workers)

    # Test data
    dataset['test'] = torch.utils.data.TensorDataset(torch.Tensor(np.array(data['test_in'])),
                                                     torch.Tensor(np.array(data['test_out'])))
    dataloaders['test'] = torch.utils.data.DataLoader(dataset['test'],
                                                      batch_size=batch_size, shuffle=shuffle,
                                                      num_workers=num_workers)
    print('Dataset Loaded')
    target_labels = list(data['target_labels'])
    train_out = np.array(data['train_out'])
    return dataloaders, target_labels, train_out

def load_single_data(path_h5, batch_size, num_workers, shuffle):
    """
    Load h5 file as a single dataset. Has to have the following fields:
        train_in : input train data
        valid_in : input validation data
        test_in : input test data
        train_out : output train labels
        valid_out : output validation labels
        test_out : output test labels
    :param path_h5: string, path to h5 file with the data
    :param batch_size: int, size of the batch in the dataloader
    :param num_workers: int, number of workers for DataLoader
    :param shuffle: boolean, shuffle parameter in DataLoader
    :return: tuple:
                DataLoader, a dataset where all train, validation, and test samples are put together
                torch.tensor, all input data
                torch.tensor, all labels of input data
    """

    data = h5py.File(path_h5, 'r')

    x = torch.Tensor(np.array(data['train_in']))
    y = torch.Tensor(np.array(data['valid_in']))
    z = torch.Tensor(np.array(data['test_in']))

    x_lab = torch.Tensor(np.array(data['train_out']))
    y_lab = torch.Tensor(np.array(data['valid_out']))
    z_lab = torch.Tensor(np.array(data['test_out']))

    res = torch.cat((x, y, z), dim=0)
    res_lab = torch.cat((x_lab, y_lab, z_lab), dim=0)

    all_dataset = torch.utils.data.TensorDataset(res, res_lab)
    dataloader  = torch.utils.data.DataLoader(all_dataset,
                                                  batch_size=128, shuffle=False,
                                                  num_workers=0)

    return dataloader, res, res_lab

def dna_one_hot(seq, seq_len=None, flatten=False):
    """
    Converts an input dna sequence to one hot encoded representation, with (A:0,C:1,G:2,T:3) alphabet

    :param seq: string, input dna sequence
    :param seq_len: int, optional, length of the string
    :param flatten: boolean, if true, makes a 1 column vector
    :return: numpy.array, one-hot encoded matrix of size (4, L), where L - the length of the input sequence
    """

    if seq_len == None:
        seq_len = len(seq)
        seq_start = 0
    else:
        if seq_len <= len(seq):
            # trim the sequence
            seq_trim = (len(seq) - seq_len) // 2
            seq = seq[seq_trim:seq_trim + seq_len]
            seq_start = 0
        else:
            seq_start = (seq_len - len(seq)) // 2

    seq = seq.upper()

    seq = seq.replace('A', '0')
    seq = seq.replace('C', '1')
    seq = seq.replace('G', '2')
    seq = seq.replace('T', '3')

    # map nt's to a matrix 4 x len(seq) of 0's and 1's.
    #  dtype='int8' fails for N's
    seq_code = np.zeros((4, seq_len), dtype='float16')
    for i in range(seq_len):
        if i < seq_start:
            seq_code[:, i] = 0.25
        else:
            try:
                seq_code[int(seq[i - seq_start]), i] = 1
            except:
                seq_code[:, i] = 0.25

    # flatten and make a column vector 1 x len(seq)
    if flatten:
        seq_code = seq_code.flatten()[None, :]

    return seq_code


def convert_onehot_back_to_seq(dataloader):
    """
    Converts one-hot encoded matrices back to DNA sequences
    :param dataloader: pytorch, DataLoader
    :return: list of strings, DNA sequences
    """

    sequences = []
    code = ['A', 'C', 'G', 'T']
    for seqs, labels in tqdm(dataloader, total=len(dataloader)):
        x = seqs.permute(0, 1, 3, 2)
        x = x.squeeze(-1)
        for i in range(x.shape[0]):
            seq = ""
            for j in range(x.shape[-1]):
                try:
                    seq = seq + code[int(np.where(x[i, :, j] == 1)[0])]
                except:
                    print("error")
                    print(x[i, :, j])
                    print(np.where(x[i, :, j] == 1))
                    break
            sequences.append(seq)
    return sequences


def _flip(x, dim):
    """
    Adapted from Selene:
    https://github.com/FunctionLab/selene/blob/master/selene_sdk/utils/non_strand_specific_module.py

    Reverses the elements in a given dimension `dim` of the Tensor.
    source: https://github.com/pytorch/pytorch/issues/229
    """

    xsize = x.size()
    dim = x.dim() + dim if dim < 0 else dim
    x = x.contiguous()
    x = x.view(-1, *xsize[dim:])
    x = x.view(
        x.size(0), x.size(1), -1)[:, getattr(
        torch.arange(x.size(1) - 1, -1, -1),
        ("cpu", "cuda")[x.is_cuda])().long(), :]

    return x.view(xsize)

def pearson_loss(x,y):
    """
    Loss that is based on Pearson correlation/objective function
    :param x: torch, input data
    :param y: torch, output labels
    :return: torch, pearson loss per sample
    """

    mx = torch.mean(x, dim=1, keepdim=True)
    my = torch.mean(y, dim=1, keepdim=True)
    xm, ym = x - mx, y - my

    cos = nn.CosineSimilarity(dim=1, eps=1e-6)
    loss = torch.sum(1-cos(xm,ym))
    return loss


def change_grad_filters(model, val, index):
    """
    Function to modify the gradient of filters of interest by the specified value
    :param model: ExplaiNN model
    :param val: float or int, value by which gradients will be multiplied
    :param index: list, gradients of which unit filters to modify
    :return: ExplaiNN model
    """

    def replace_val(grad, val, index):
        grad[index, :, :] *= val

        return grad

    model.linears[0].weight.register_hook(lambda grad: replace_val(grad, val, index))

    return model

def _PWM_to_filter_weights(pwm, filter_size=19):
    """
    Function to convert a given pwm into convolutional filter
    :param pwm: list of length L (size of the PWM) with lists of size 4 (nucleotide values)
    :param filter_size: int, the size of the pwm
    :return: numpy.array, of shape (L, 4)
    """

    # Initialize
    lpop = 0
    rpop = 0

    pwm = [[.25,.25,.25,.25]]*filter_size+pwm+[[.25,.25,.25,.25]]*filter_size

    while len(pwm) > filter_size:
        if max(pwm[0]) < max(pwm[-1]):
            pwm.pop(0)
            lpop += 1
        elif max(pwm[-1]) < max(pwm[0]):
            pwm.pop(-1)
            rpop += 1
        else:
            if lpop > rpop:
                pwm.pop(-1)
                rpop += 1
            else:
                pwm.pop(0)
                lpop += 1

    return(np.array(pwm) - .25)

def read_meme(meme_file):
    """
    Function the motifs from the input meme file
    Right now works only if motifs are of the same length
    :param meme_file: string, path to the meme file
    :return: tuple:
                numpy.array, matrix of pwms of size (N, L, 4), where N is the number of motifs, L is the motif length,
                4 number of nucleotides
    """

    with open(meme_file) as fp:
        line = fp.readline()
        motifs = []
        motif_names = []
        while line:
            # determine length of next motif
            if line.split(" ")[0] == 'MOTIF':
                line = line.strip()
                # add motif number to separate array
                motif_names.append(line.split(" ")[1])
                # get length of motif
                line2 = fp.readline().split(" ")
                motif_length = int(float(line2[5]))
                # read in motif
                current_motif = np.zeros((motif_length, 4))  # Edited pad shorter ones with 0
                for i in range(motif_length):
                    current_motif[i, :] = fp.readline().strip().split()
                motifs.append(current_motif)
            line = fp.readline()
        motifs = np.stack(motifs)
        motif_names = np.stack(motif_names)

    return motifs, motif_names


"""
ExplaiNN
interpretable and transparent neural networks for genomics
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

# The directory containing this file
# HERE = path.abspath(path.dirname(__file__))
#
# # Get the long description from the README file
# with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
#     long_description = f.read()

long_description = "interpretable and transparent neural networks for genomics"

setup(
    name="explainn",
    version="0.1.5",
    description=long_description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Gherman Novakovsky @ wassermanlab",
    author_email="gnovakovsky@cmmt.ubc.ca",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent"
    ],
    packages=find_packages(), # ["explainn"],
    include_package_data=True,
    install_requires=["numpy", "h5py", "tqdm", "pandas"]
)